/**
 * Computación Paralela
 *
 *
 * Práctica Stencil - Versión secuencial base
 */
// Muestra un display con la matrix (para matrices pequeñas)
// Compilar: make display
//#define SHOW_DISPLAY

// Bibliotecas estándard
#include <time.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

// Biblioteca de funciones para computacion paralela
#include "cputils.cu"


// Kernels CUDA
#include "kernel.cu"


// Biblioteca de CUDA
#include <cuda.h>

// Constantes de los pesos del stencil
#define WEIGHT_CENTER   (float)  (0.5   * (1.0/1.0) )
#define WEIGHT_ADJACENT (float) (0.295 * (1.0/4.0) )
#define WEIGHT_CORNER   (float) (0.195 * (1.0/4.0) )

// Máximo residuo permitido
#define MAX_RESID       (float) 0.001

// Número máximo de hilos por bloque para Pre Fermi
#define MAX_BLQ_DIM	(int) 512

// Tamaño de cada matriz
#define MAT000	(int) 8*8
#define MAT00	(int) 16*16
#define MAT0	(int) 32*32
#define MAT1	(int) 256*256
#define MAT2	(int) 1024*1024
#define MAT3	(int) 2048*2048
#define MAT4	(int) 4096*4096
#define MAT5	(int) 5120*5120

// Númeor de hilos límites definidos para X e Y
#define X_DIV	(int) 32
#define Y_DIV	(int) 16 

/**
 * Suma los valores de matrix para verificar la solución
 * @param matrix La matrix
 * @param elems El número de elementos.
 */
m_type check_sum(m_type * matrix, int elems);

/**
 * Función main
 */
int main(int argc, char *argv[]){

	//************************************************************
	// 0. Variables
	char * matrix_name; /**< Nombre de la matrix de entrada */
	int rows;           /**< Número de filas de la matriz */
	int cols;           /**< Número de columnas */
	m_type * h_matrix1;   /**< Matrix con los datos */
	m_type * d_matrix1;  /**< Matrix con los datos que va a DEVICE*/
	m_type * d_matrix2;  /**< Matrix con en la a que se actualiza en DEVICE*/
	m_type * d_matrixAux;  /**< Matrix auxiliar para realizar el swap de las matrices de DEVICE*/
	double t_begin;     /**< Tiempo de inicio del calculo */
	double t_end;       /**< Tiempo de finalización */
	m_type sum;         /**< Suma de verificación */
	int bX; /**< Numero de columnas del bloque*/
	int bY; /**< Numero de filas del bloque*/
	int gX; /**< Numero de columnas del grid*/
	int gY; /**< Numero de filas del grid*/
	int matrix_size; /**< Tamanyo de la matriz leida*/
	int iter = 0;	/**< Iteraciones realizadas hasta llegar a un punto estable */
	/* Variables para el cálculo de residuos.  */
	m_type resid; /**< Residuo en HOST*/
	m_type * d_resid; /**< Donde devuelve el residuo el kernel de update*/
	// Macros para acceder a las matrices
#define h_m1(i,j) (h_matrix1[(i)*(cols)+(j)])
#define h_m2(i,j) (h_matrix2[(i)*(cols)+(j)])
	//************************************************************
	
	//************************************************************
	// 1. Leer los parametros de entrada.
	if(argc != 2){
		fprintf(stderr,"USO: %s <matriz de datos>\n",argv[0]);
		exit(EXIT_FAILURE);
	}
	
	matrix_name = argv[1];		
	//************************************************************

	
	//************************************************************
	// 2. Cargar la matriz
	// Matrix de tamaño rows+2 x cols+2 (halo de tamaño 1)
	// (Los datos están en [1,rows+1][1,cols+1])
	//************************************************************
	// 2. Cargar la matriz
	// Matrix de tamaño rows+2 x cols+2 (halo de tamaño 1)
	// (Los datos están en [1,rows+1][1,cols+1])
	cp_read_matrix_size(matrix_name, &rows, &cols);
	matrix_size = rows * cols;
	// Definición de geometría
	// CASO BASE
	if (matrix_size<=MAX_BLQ_DIM){
		bX=rows;	
		bY=cols;	
		gX=1;	
		gY=1;	
	} else if (((rows%X_DIV)!=0)||((cols%Y_DIV)!=0)){
		fprintf(stderr,"El programa requiere matrices cuyos número de columnas sea múltiplos de %d y sus filas múltiplo de %d. La matriz dada es de %dx%d.\n",X_DIV,Y_DIV,rows,cols);
		exit(EXIT_FAILURE);
	// CASO AVANZADO (Hasta matrices 5120x5120)
	} else {
		bX=X_DIV;	
		bY=Y_DIV;	
		gX=rows/X_DIV;	
		gY=cols/Y_DIV;	
 	}

	printf(" Matriz %s: %dx%d\n",matrix_name, rows, cols);

	//Reserva para variables de HOST
	h_matrix1 = (float*) malloc(sizeof(m_type) * (size_t) (matrix_size));

	//Reserva para variables de DEVICE
	cudaMalloc((void**) &d_matrix1, sizeof(m_type) * (int) (matrix_size));
	cudaMalloc((void**) &d_matrix2, sizeof(m_type) * (int) (matrix_size));
	cudaMalloc((void**) &d_resid, sizeof(m_type) );//rows * cols));

	cp_read_matrix(matrix_name, h_matrix1, 0);

	dim3 bloqShapeGpuFunc(bX,bY); 
	dim3 gridShapeGpuFunc(gX,gY);

#ifdef SHOW_DISPLAY
	cp_display_create("Stencil", rows+2, cols+2);
	cp_display_draw_matrix(h_matrix1,CP_RGB(255,0,0),CP_RGB(0,255,0));
	cp_msleep(1000);
#endif

	
	//************************************************************
	// 4. Bucle principal
	t_begin = cp_Wtime();
	cudaMemcpy(d_matrix1, h_matrix1, sizeof(m_type) * (matrix_size), cudaMemcpyHostToDevice);
	do {
		
		resid = 0.0;

		cudaMemcpy(d_resid, &resid, sizeof(m_type) , cudaMemcpyHostToDevice);
		gpuFunc_Update<<<gridShapeGpuFunc,bloqShapeGpuFunc>>>(d_matrix1,d_matrix2,rows,cols,d_resid);
		cudaMemcpy(&resid, d_resid, sizeof(m_type) , cudaMemcpyDeviceToHost);

		d_matrixAux = d_matrix1;
		d_matrix1 = d_matrix2;
		d_matrix2 = d_matrixAux;
		
#ifdef SHOW_DISPLAY
		cp_display_draw_matrix(h_matrix2,CP_RGB(255,0,0),CP_RGB(0,255,0));
		cp_msleep(50); // Para ver más despacio la evolucion del proceso
#endif

		iter++;

	} while (resid > MAX_RESID);
	//************************************************************
	
	//************************************************************
	// 5. Suma de verificación
	int tmp_matrix_size = matrix_size;
	if (matrix_size>MAT0){
		while(tmp_matrix_size>MAX_BLQ_DIM){
			tmp_matrix_size/=MAX_BLQ_DIM;
			gpuFunc_Reduce<<<tmp_matrix_size,MAX_BLQ_DIM,MAX_BLQ_DIM*sizeof(m_type)>>>(d_matrix1,d_matrix1);
		}
	}
	cudaMemcpy(h_matrix1,d_matrix1,(tmp_matrix_size)*sizeof(m_type),cudaMemcpyDeviceToHost);	
	sum=check_sum(h_matrix1,tmp_matrix_size);

	t_end = cp_Wtime();
	//************************************************************
	
	
	//************************************************************
	// 6. Mostrar resultados
	printf(" Check sum: %f\n", sum);
	printf(" Iteraciones: %d\n", iter);
	printf(" Tiempo de ejecución: %f\n", t_end-t_begin);
	//************************************************************

	//************************************************************

#ifdef SHOW_DISPLAY	
	cp_msleep(250);
	cp_display_close();
#endif


	//************************************************************
	// 7. Liberar las matrices
	free(h_matrix1);
	cudaFree(d_matrix1);
	cudaFree(d_matrix2);
	cudaFree(d_resid);
	cudaThreadExit();
	//************************************************************

#undef h_m1
#undef h_m2
	return EXIT_SUCCESS;
}


m_type check_sum(m_type * matrix, int elems){
	m_type sum = 0.0;
	int i;
	for(i=0; i<elems; i++)
			sum += matrix[i];

	return sum;
}
