/**
 * Computación Paralela
 *
 *
 * Práctica Stencil - Versión secuencial base
 */

// Muestra un display con la matrix (para matrices pequeñas)
// Compilar: make display
// #define SHOW_DISPLAY

// Bibliotecas estándard
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

// Biblioteca de funciones para computacion paralela
#include <cputils.h>

// Biblioteca MPI 
#include <mpi.h>

// Constantes de los pesos del stencil
#define WEIGHT_CENTER   (0.5   * (1.0/1.0) )
#define WEIGHT_ADJACENT (0.295 * (1.0/4.0) )
#define WEIGHT_CORNER   (0.195 * (1.0/4.0) )
#define MAX_RESID 0.001

// Constantes definidas para el tipo MPI_Cols
#define NDIMS 2 // Dimensión de las matrices

// Constantes definidas para el paso de halos
#define SUP_SND 0 // Dimensión de las matrices
#define INF_SND 1 // Dimensión de las matrices
#define SUP_RCV 2 // Dimensión de las matrices
#define INF_RCV 3 // Dimensión de las matrices

/** Variable para el cálculo de residuos.  */
m_type residParcial;

/** Buffer auxiliar añadido por tener la funcion cp_matrix2string_m_type() */
// Aux buffer
char print_buffer[1024];


/**
* Convierte una matrix de floats a cadaena de caracteres
* @param matrix La matrix a convertir
* @param rows El numero de filas de la matrix
* @param cols El numero de columnas de la matrix
*/
char * cp_matrix2string_m_type(float * matrix, int rows, int cols);

/**
*  Genera la siguiente iteración
* @param matrix_old El estado anterior.
* @param matrix_new El nuevo estado.
* @param rows El número de filas.
* @param cols El número de columnas.
*/
void update(m_type * matrix_old, m_type * matrix_new, int rows, int cols);

/**
*  Envía las submatrices al resto de procesadores desde 0
* @param matrix La matrix a partir.
* @param rows El número de filas.
* @param cols El número de columnas.
* @param tag El tag de la comunicacion.
* @param size El El numero de procesadores en la comunicacion.
*/
void send_sub_matrix(m_type * matrix, int rows, int cols, int tag, int size);

/**
*  Recibe las submatrices al resto de procesadores desde 0
* @param matrix La matrix de recepcion.
* @param rows El número de filas.
* @param cols El número de columnas.
* @param tag El tag de la comunicacion.
* @param size El El numero de procesadores en la comunicacion.
*/
void recv_sub_matrix(m_type * matrix, int rows, int cols, int tag, int size);

/**
*  Envía el halo las submatrices desde el procesador i al procesador (i+1)%size y al (i-1+size)%size segun corresponda.
* @param matrix La matrix de envio.
* @param halos Puntero al array de halos (superior de envío, inferior de envio, superior de recepcion e inferior de recepcion).
* @param rows El número de filas.
* @param cols El número de columnas.
* @param tag El tag de la comunicacion.
* @param size El numero de procesadores en la comunicacion.
* @param rank El numero de procesador que comienza la comunicacion.
* @param reqs Puntero al array de request de espera de envio del halo superior, envio halo inferior y recepcion de halo inferior.
*/
void send_recv_halos(m_type * matrix, m_type * halos, int rows, int cols, int tag, int size, int rank, MPI_Request * reqs);

/**
 * Suma los valores de matrix para verificar la solución
 * @param matrix La matrix
 * @param rows El número de filas.
 * @param cols El número de columnas.
 */
m_type check_sum(m_type * matrix, int rows, int cols);



/**
 * Función main
 */
int main(int argc, char *argv[]){

        //************************************************************
        // 0. Variables
        char * matrix_name; /**< Nombre de la matrix de entrada */
        int rows;           /**< Número de filas de la matriz */
        int cols;           /**< Número de columnas */
        m_type * matrix1 = NULL;   /**< Matrix con los datos */
	m_type * matrix2;   /**< Submatrix de paso */
        m_type * matrix3;   /**< Copia de la matrix para calcular nuevos valores */
	m_type * aux;	    /**< Matrix auxiliar para swap entre matrix1 a matrix3 */
        double t_begin;     /**< Tiempo de inicio del calculo */
        double t_end;       /**< Tiempo de finalización */
	double t_rank_max;  /**< Tiempo máximo de ejecucion */
	double t_rank;      /**< Tiempo de ejecucion del rank*/
        m_type sum;         /**< Suma de verificación */
        int iter = 0;           /**< Iteraciones realizadas hasta llegar a un punto estable */
	int rank, size;     /**< Variables para MPI*/
	int tag = 0;     /**< tag por defecto para MPI*/
        // Macros para acceder a las matrices
#define m1(i,j) (matrix1[(i)*(cols)+(j)])
#define m2(i,j) (matrix2[(i)*(cols)+(j)])
        //************************************************************
	
        //************************************************************
        // 1. Iniciar MPI
        MPI_Init( &argc, &argv );
        MPI_Comm_rank( MPI_COMM_WORLD, &rank );
        MPI_Comm_size( MPI_COMM_WORLD, &size );
        //************************************************************


       //************************************************************
        // 2. Leer los parametros de entrada.
        if(argc != 2){
                fprintf(stderr,"USO: %s <matriz de datos>\n",argv[0]);
                exit(EXIT_FAILURE);
        }

        matrix_name = argv[1];
        //************************************************************


        //************************************************************
        // 3. Cargar la matriz
        // Matrix de tamaño rows+2 x cols+2 (halo de tamaño 1)
        // (Los datos están en [1,rows+1][1,cols+1])
        cp_read_matrix_size(matrix_name, &rows, &cols);
	if (rank == 0){
	        printf(" Matriz %s: %dx%d\n",matrix_name, rows, cols);
        // int matrix_size = (rows + 2) * (cols + 2);
	// No vamos a usar halo
        	int matrix_size = rows * cols;
	        matrix1 = malloc(sizeof(m_type) * (size_t) matrix_size);
        // cp_read_matrix(matrix_name, matrix1, 1);
	// No vamos a usar halo
        	cp_read_matrix(matrix_name, matrix1, 0);
	//}
        //************************************************************

//BORRAR	printf("Matriz origen: %s\n",cp_matrix2string_m_type(matrix1, rows, cols));
	}


        //************************************************************
        // 5. Enviamos la submatrix  
	// Envío
	int sub_rows;
	if (rank == 0){
		send_sub_matrix(matrix1, rows, cols, tag, size);
		int i, j;
		
		sub_rows=rows/size;
        
		matrix2 = malloc(sizeof(m_type) * (size_t) (sub_rows+2)*(cols));

		for (i=0;i<sub_rows;i++){
			for (j=0;j<cols;j++){
				m2(i+1,j)=m1(i,j);
			}
		}
		
	}		
        //************************************************************

        //************************************************************
        // 6. Recibimos la submatrix  

	// Recepción
	if (rank != 0){
//BORRAR		printf("Soy %d\n",rank);

		// Se fija el tamaño de las submatrices según sea el último procesador o no
		if (rank == size-1){
			sub_rows = rows/size + rows%size; // Ultimo procesador
		} else {
			sub_rows = rows/size; // Para el resto
		} 

		matrix2 = malloc(sizeof(m_type) * (size_t) ((sub_rows+2)*cols));

		recv_sub_matrix(matrix2, sub_rows, cols, tag, size);
	}
        //************************************************************

        //************************************************************
        // 7. Envio de halos  

	m_type * halos = malloc (sizeof(m_type) * (size_t) (cols*4));
	MPI_Request haloReq[3];

	send_recv_halos(matrix2, halos, sub_rows, cols, tag, size, rank, haloReq);
        //************************************************************

	//printf("[RANK %d]Matriz destino: %s\n",rank,cp_matrix2string_m_type(matrix2, sub_rows+2, cols));

        //************************************************************
        // 9. Calculo de nueva matrix. update();
	matrix3 = malloc(sizeof(m_type) * (size_t) (sub_rows+2)*(cols));
	m_type residMax;

	t_begin = MPI_Wtime();

	do {

		residParcial = 0.0;

		update(matrix2,matrix3,sub_rows,cols);

	// 10. Swap de matrices
		aux = matrix2;
		matrix2 = matrix3;
		matrix3 = aux;
	
		send_recv_halos(matrix2, halos, sub_rows, cols, tag, size, rank, haloReq);
	
		//printf("[RANK %d]Matriz destino MOD: %s\n\tResiduo: %f\n",rank,cp_matrix2string_m_type(matrix2, sub_rows+2, cols),residParcial);

		MPI_Allreduce(&residParcial,&residMax,1,MPI_FLOAT,MPI_MAX,MPI_COMM_WORLD);
		residParcial = residMax;
		
		//printf("[RANK %d] Residuo: %f\n",rank,residMax);
		iter++;
//		if (iter == 6){
//		
//			printf("[RANK %d]Matriz destino MOD: %s\n\tResiduo: %f\n",rank,cp_matrix2string_m_type(matrix2, sub_rows+2, cols),residParcial);
//		}
	} while (residParcial > MAX_RESID);
//	} while (iter < 1);
        //************************************************************

//	printf("[RANK %d]Matriz destino MOD: %s\n\tResiduo: %f\n",rank,cp_matrix2string_m_type(matrix2, sub_rows+2, cols),residParcial);

        //************************************************************
        // 11. Suma de verificacion
	m_type sum_parcial = check_sum(matrix2,sub_rows+2,cols);
	MPI_Allreduce(&sum_parcial,&sum,1,MPI_FLOAT,MPI_SUM,MPI_COMM_WORLD);
        //************************************************************

        //************************************************************
        // 12. Toma de tiempo
	t_end = MPI_Wtime();
	t_rank = t_end - t_begin; 
	MPI_Reduce(&t_rank,&t_rank_max,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
        //************************************************************

        //************************************************************
        // 6. Mostrar resultados
	if (rank == 0){
       		printf(" Check sum: %f\n", sum);
	        printf(" Iteraciones: %d\n", iter);
        	printf(" Tiempo de ejecución: %f\n", t_rank_max);
	}
        //************************************************************


	//printf("[%d]\n%s\n\n",rank,cp_matrix2string_m_type(matrix2,sub_rows,cols));
	
        //************************************************************
        // 11. Liberar las matrices
	if (rank == 0){
        	free(matrix1);
	}
        free(matrix2);
        free(matrix3);
	free(halos);
        //************************************************************

#undef m1
#undef m2
//BORRAR	printf("Yo %d finalizo\n",rank);
	MPI_Finalize();	
        return EXIT_SUCCESS;
}

char * cp_matrix2string_m_type(float * matrix, int rows, int cols){ 
 
        int i,j; 
        int nchars; 
        char * buffer = print_buffer; 
 
 
        nchars = sprintf(buffer,"\n"); 
        buffer += nchars; 
 
        for(i=0; i<rows; i++){ 
 
                nchars = sprintf(buffer,"["); 
                buffer += nchars; 
 
                for(j=0; j<cols-1; j++){ 
                        nchars = sprintf(buffer,"%6.4f,",matrix[cols*i+j]); 
                        buffer += nchars; 
                } 
 
                nchars = sprintf(buffer,"%6.4f]\n",matrix[cols*i+j]);
                buffer += nchars;
        }

        return print_buffer;
}

void update(m_type * matrix_old, m_type * matrix_new, int rows, int cols){
	//************************************************************
        // 1. Macros para acceder a las matrices
        #define m_old(i,j) (matrix_old[(i)*(cols)+(j)])
        #define m_new(i,j) (matrix_new[(i)*(cols)+(j)])
        //************************************************************

        //************************************************************
        // 2. Bucle principal
	int i, j;
	m_type tmpResid;
	for(i=1;i<rows+1;i++){
		for(j=0;j<cols;j++){
			m_new(i,j) = (m_old(i,j) * WEIGHT_CENTER) +
					((m_old(i-1,j) + 
						m_old(i+1,j) +
						m_old(i,(j-1+cols)%cols) +
						m_old(i,(j+1)%cols)) * WEIGHT_ADJACENT) +
					((m_old(i-1,(j-1+cols)%cols) +
						m_old(i-1,(j+1)%cols) +
						m_old(i+1,(j-1+cols)%cols) +
						m_old(i+1,(j+1)%cols)) * WEIGHT_CORNER);

			tmpResid = fabs (m_old(i,j) - m_new(i,j));
			//printf("[%d]%f\n",rank,tmpResid);
			if (tmpResid > residParcial) residParcial = tmpResid;

		}
	}
        //************************************************************

        #undef m_old
        #undef m_new
}

void send_sub_matrix(m_type * matrix, int rows, int cols, int tag, int size){
        //************************************************************
	// 0. Creamos el tipo para enviar cada trozo.
        MPI_Datatype MPI_Col;

        int array_of_sizes[NDIMS]; // Tamaño de las matrices 
        int array_of_subsizes[NDIMS];// Tamaño de las sub matrices
        int array_of_starts[NDIMS]; // Comienzos de las matrices
        int order = MPI_ORDER_C; // row-wise or column-wise 
        //************************************************************


        //************************************************************
	// 1. Creamos el tipo para enviar cada trozo.
		array_of_starts[1] = 0;

		array_of_sizes[0] = rows;
		array_of_sizes[1] = cols;

		array_of_subsizes[1] = cols;

		int i;
		for (i=1;i<size;i++){
			array_of_starts[0] = (rows/size)*i;

			if (i != size-1){
				array_of_subsizes[0] = rows/size;
			} else {
				array_of_subsizes[0] = rows/size + rows%size;
			}
			
			MPI_Type_create_subarray(NDIMS,array_of_sizes,array_of_subsizes,array_of_starts,order,MPI_INT,&MPI_Col);
        		MPI_Type_commit(&MPI_Col);
        //************************************************************

        //************************************************************
	// 2. Enviamos cada trozo.
                	// Envio 
	                MPI_Send(matrix,1,MPI_Col,i,tag,MPI_COMM_WORLD);
//BORRAR			printf("Trozo enviado a %d\n",i);

	        	MPI_Type_free(&MPI_Col);
        //************************************************************
		}
	
}


void recv_sub_matrix(m_type * matrix, int rows, int cols, int tag, int size){
        //************************************************************
	// 0. Creamos el tipo para enviar cada trozo.
        MPI_Datatype MPI_Col;

        int array_of_sizes[NDIMS]; // Tamaño de las matrices 
        int array_of_subsizes[NDIMS];// Tamaño de las sub matrices
        int array_of_starts[NDIMS]; // Comienzos de las matrices
        int order = MPI_ORDER_C; // row-wise or column-wise 
        //************************************************************

//BORRAR		printf("Soy %d\n",rank);


	array_of_starts[0] = 1;
	array_of_starts[1] = 0;

	array_of_sizes[0] = rows+2;
	array_of_sizes[1] = cols;

	array_of_subsizes[0] = rows;
	array_of_subsizes[1] = array_of_sizes[1];

	MPI_Type_create_subarray(NDIMS,array_of_sizes,array_of_subsizes,array_of_starts,order,MPI_INT,&MPI_Col);
       	MPI_Type_commit(&MPI_Col);

//BORRAR		printf("Soy %d recibiendo\n",rank);
	MPI_Recv(matrix,1,MPI_Col,0,tag,MPI_COMM_WORLD,MPI_STATUS_IGNORE);

       	MPI_Type_free(&MPI_Col);
//BORRAR		printf("Era %d \n",rank);
}

//void send_recv_halos(matrix3, halos, sub_rows, cols, tag, size, rank, &haloReq){
void send_recv_halos(m_type * matrix, m_type * halos, int rows, int cols, int tag, int size, int rank, MPI_Request * reqs){
	//************************************************************
        // 0. Macros para acceder a las matrices
        #define m(i,j) (matrix[(i)*(cols)+(j)])
        #define h(i,j) (halos[(i)*(cols)+(j)])
        //************************************************************

	//MPI_Request supSndReq, infSndReq, infRcvReq;

	int i;
	for (i=0;i<cols;i++){
		//halo_sup[i]=m(1,i);//matrix3[cols+i];
		h(SUP_SND,i)=m(1,i);//matrix3[cols+i];
		//halo_inf[i]=m(rows,i);//matrix3[rows/size+i];
		h(INF_SND,i)=m(rows,i);//matrix3[rows/size+i];
	}
//BORRAR		printf("[RANK %d]Halo superior: %s\n",rank,cp_matrix2string_m_type(halo_supS, 1, cols));
//BORRAR		printf("[RANK %d]Halo inferior: %s\n",rank,cp_matrix2string_m_type(halo_infS, 1, cols));

	MPI_Isend(halos+(cols*SUP_SND),cols,MPI_INT,(rank-1+size)%size,tag,MPI_COMM_WORLD,reqs+SUP_SND);
	MPI_Isend(halos+(cols*INF_SND),cols,MPI_INT,(rank+1)%size,tag,MPI_COMM_WORLD,reqs+INF_SND);
	MPI_Irecv(halos+(cols*INF_RCV),cols,MPI_INT,(rank+1)%size,tag,MPI_COMM_WORLD,reqs+INF_RCV);
	MPI_Recv(halos+(cols*SUP_RCV),cols,MPI_INT,(rank-1+size)%size,tag,MPI_COMM_WORLD,MPI_STATUS_IGNORE);

	MPI_Wait(reqs+SUP_SND,MPI_STATUS_IGNORE);
	MPI_Wait(reqs+INF_SND,MPI_STATUS_IGNORE);
	MPI_Wait(reqs+INF_RCV,MPI_STATUS_IGNORE);

	for(i=0;i<cols;i++){
		//m(0,i)=halo_supR[i];
		m(0,i)=h(SUP_RCV,i);
		//m(sub_rows+1,i)=halo_infR[i];
		m(rows+1,i)=h(INF_RCV,i);
	}
	#undef m
	#undef h
}


m_type check_sum(m_type * matrix, int rows, int cols){

        //************************************************************
        // 1. Macros para acceder a la matriz
        #define m(i,j) (matrix[(i)*(cols)+(j)])
        //************************************************************

        //************************************************************
        // 2. Bucle con la suma
        m_type sum = 0;
        int i,j;
        for(i=1; i<rows-1; i++)
                for(j=0; j<cols; j++)
                        sum += m(i,j);
        //************************************************************

        return sum;

        #undef m
}

